@@ -9,10 +9,19 @@ import sys
 import os
 import numpy as np
 import matplotlib.pyplot as plt
+from mpl_toolkits.axes_grid1.inset_locator import inset_axes
+import seaborn as sns
+
+sns.set_style("white")
 import scanpy as sc
 import time
 import threading
-from skimage.filters import threshold_li, threshold_otsu, threshold_mean
+from skimage.filters import (
+    threshold_li,
+    threshold_otsu,
+    threshold_mean,
+    threshold_multiotsu,
+)
 
 from .logistic import LogitNet
 
@@ -36,15 +45,15 @@ def recipe_dropkick(
         adata (AnnData.AnnData): object with raw counts data in .X
         filter (bool): remove cells with less than min_genes detected
             and genes with zero total counts
-        min_genes (int): threshold for minimum genes detected. Default 50.
+        min_genes (int): threshold for minimum genes detected.
             Ignored if filter==False.
         calc_metrics (bool): if False, do not calculate metrics in .obs/.var
         mito_names (str): substring encompassing mitochondrial gene names for
             calculation of mito expression. Ignored if calc_metrics==False.
-        n_ambient (int): number of ambient genes to call. top genes by cells.
+        n_ambient (int): number of ambient genes to call (top genes by cells).
             Ignored if calc_metrics==False.
         target_sum (int): total sum of counts for each cell prior to arcsinh 
-            or log1p transformations; default None to use median counts.
+            or log1p transformations; if None, use median counts.
         n_hvgs (int or None): number of HVGs to calculate using Seurat method.
             if None, do not calculate HVGs.
         X_final (str): which normalization layer should be left in .X slot?
@@ -53,7 +62,7 @@ def recipe_dropkick(
 
     Returns:
         adata (AnnData.AnnData): updated object includes:
-            - useful .obs and .var columns
+            - useful .obs and .var columns (if calc_metrics==True)
                 ("total_counts", "pct_counts_mito", "n_genes_by_counts", etc.)
             - raw counts (adata.layers["raw_counts"])
             - arcsinh-transformed normalized counts (adata.layers["arcsinh_norm"])
@@ -67,7 +76,7 @@ def recipe_dropkick(
         if verbose:
             if adata.shape[0] != orig_shape[0]:
                 print(
-                    "Ignoring {} cells with less than {} genes detected".format(
+                    "Ignoring {} barcodes with less than {} genes detected".format(
                         orig_shape[0] - adata.shape[0], min_genes
                     )
                 )
@@ -128,13 +137,6 @@ def recipe_dropkick(
         adata.obs["arcsinh_n_genes_by_counts"] = np.arcsinh(
             adata.obs["n_genes_by_counts"]
         )
-        # make sure everything went well...
-        np.testing.assert_approx_equal(
-            adata.var.pct_dropout_by_counts.min(),
-            lowest_dropout,
-            significant=7,
-            verbose=True,
-        )
 
     # normalize counts before transforming
     sc.pp.normalize_total(adata, target_sum=target_sum, layers=None, layer_norm=None)
@@ -172,43 +174,45 @@ def recipe_dropkick(
 def auto_thresh_obs(
     adata,
     obs_cols=["arcsinh_n_genes_by_counts", "pct_counts_ambient"],
-    # double=["arcsinh_n_genes_by_counts"],
-    # single=["pct_counts_ambient"],
-    method="otsu",
+    methods=["multiotsu", "otsu"],
 ):
     """
     automated thresholding on metrics in adata.obs
 
     Parameters:
         adata (anndata.AnnData): object containing unfiltered scRNA-seq data
-        double (list of str): name of column(s) to double threshold from adata.obs
-        single (list of str): name of column(s) to single threshold from adata.obs
-        method (str): one of 'otsu' (default), 'li', or 'mean'
+        obs_cols (list of str): name of column(s) to threshold from adata.obs
+        methods (list of str): one of 'otsu', 'multiotsu', 'li', or 'mean'
 
     Returns:
         thresholds (dict): keys are obs_cols and values are threshold results
     """
-    thresholds = dict.fromkeys(obs_cols)  # initiate output dictionary
-    # for col in double:
-    #    tmp = np.array(adata.obs[col])
-    #    thresholds[col] = threshold_multiotsu(tmp)
-    for col in obs_cols:
-        tmp = np.array(adata.obs[col])
-        if method == "otsu":
-            thresholds[col] = threshold_otsu(tmp)
-        elif method == "li":
-            thresholds[col] = threshold_li(tmp)
-        elif method == "mean":
-            thresholds[col] = threshold_mean(tmp)
+    # convert to lists before looping
+    if isinstance(obs_cols, str):
+        obs_cols = [obs_cols]
+    if isinstance(methods, str):
+        methods = [methods]
+    # initiate output dictionary
+    thresholds = dict.fromkeys(obs_cols)
+    for i in range(len(obs_cols)):
+        tmp = np.array(adata.obs[obs_cols[i]])
+        if methods[i] == "multiotsu":
+            thresholds[obs_cols[i]] = threshold_multiotsu(tmp)
+        elif methods[i] == "otsu":
+            thresholds[obs_cols[i]] = threshold_otsu(tmp)
+        elif methods[i] == "li":
+            thresholds[obs_cols[i]] = threshold_li(tmp)
+        elif methods[i] == "mean":
+            thresholds[obs_cols[i]] = threshold_mean(tmp)
         else:
             raise ValueError(
-                "Please provide a valid threshold method ('otsu', 'li', 'mean')."
+                "Please provide a valid threshold method ('otsu', 'multiotsu', 'li', 'mean')."
             )
 
     return thresholds
 
 
-def plot_thresh_obs(adata, thresholds, bins=40, show=True):
+def plot_thresh_obs(adata, thresholds, bins=40, save_to=None, verbose=True):
     """
     plot automated thresholding on metrics in adata.obs as output by auto_thresh_obs()
 
@@ -216,25 +220,39 @@ def plot_thresh_obs(adata, thresholds, bins=40, show=True):
         adata (anndata.AnnData): object containing unfiltered scRNA-seq data
         thresholds (dict): output of auto_thresh_obs() function
         bins (int): number of bins for histogram
-        show (bool): show plot or return object
+        save_to (str): path to .png file for saving figure; returns figure by default
+        verbose (bool): print updates to console
 
     Returns:
-        plot of distributions of obs_cols in thresholds dictionary with corresponding threshold values
+        plot of distributions of obs_cols in thresholds dictionary with corresponding thresholds
     """
     fig, axes = plt.subplots(
         ncols=len(thresholds), nrows=1, figsize=(len(thresholds) * 4, 4), sharey=True
     )
-    axes[0].set_ylabel("cells")
-    for i in range(len(thresholds)):
-        axes[i].hist(adata.obs[list(thresholds.keys())[i]], bins=bins)
-        if isinstance(list(thresholds.values())[i], np.ndarray):
-            [axes[i].axvline(_x, color="r") for _x in list(thresholds.values())[i]]
+    # if multiple plots, loop through axes
+    if len(thresholds) > 1:
+        axes[0].set_ylabel("cells")
+        for i in range(len(thresholds)):
+            axes[i].hist(adata.obs[list(thresholds.keys())[i]], bins=bins)
+            if isinstance(list(thresholds.values())[i], np.ndarray):
+                [axes[i].axvline(_x, color="r") for _x in list(thresholds.values())[i]]
+            else:
+                axes[i].axvline(list(thresholds.values())[i], color="r")
+            axes[i].set_title(list(thresholds.keys())[i])
+    # if single plot, only one set of axes in subplot
+    else:
+        axes.set_ylabel("cells")
+        axes.hist(adata.obs[list(thresholds.keys())[0]], bins=bins)
+        if isinstance(list(thresholds.values())[0], np.ndarray):
+            [axes.axvline(_x, color="r") for _x in list(thresholds.values())[0]]
         else:
-            axes[i].axvline(list(thresholds.values())[i], color="r")
-        axes[i].set_title(list(thresholds.keys())[i])
+            axes.axvline(list(thresholds.values())[0], color="r")
+        axes.set_title(list(thresholds.keys())[0])
     fig.tight_layout()
-    if show:
-        plt.show()
+    if save_to is not None:
+        if verbose:
+            print("Saving threshold plot to {}".format(save_to))
+        fig.savefig(save_to)
     else:
         return fig
 
@@ -246,6 +264,7 @@ def filter_thresh_obs(
     directions=["above", "below"],
     inclusive=True,
     name="thresh_filter",
+    verbose=True,
 ):
     """
     filter cells by thresholding on metrics in adata.obs as output by auto_thresh_obs()
@@ -254,43 +273,84 @@ def filter_thresh_obs(
         adata (anndata.AnnData): object containing unfiltered scRNA-seq data
         thresholds (dict): output of auto_thresh_obs() function
         obs_cols (list of str): name of column(s) to threshold from adata.obs
-        directions (list of str): 'below' or 'above', indicating which direction to keep (label=1)
-        inclusive (bool): include cells at the thresholds? default True.
+        directions (list of str): 'below' or 'above', indicating which direction to keep
+        inclusive (bool): include cells at the thresholds?
         name (str): name of .obs col containing final labels
+        verbose (bool): print updates to the console
 
     Returns:
         updated adata with filter labels in adata.obs[name]
     """
+    # convert to lists before looping
+    if isinstance(obs_cols, str):
+        obs_cols = [obs_cols]
+    if isinstance(directions, str):
+        directions = [directions]
     # initialize .obs column as all "good" cells
     adata.obs[name] = 1
     # if any criteria are NOT met, label cells "bad"
     for i in range(len(obs_cols)):
-        if directions[i] == "above":
-            if inclusive:
-                adata.obs.loc[
-                    (adata.obs[name] == 1)
-                    & (adata.obs[obs_cols[i]] <= thresholds[obs_cols[i]]),
-                    name,
-                ] = 0
-            else:
-                adata.obs.loc[
-                    (adata.obs[name] == 1)
-                    & (adata.obs[obs_cols[i]] < thresholds[obs_cols[i]]),
-                    name,
-                ] = 0
-        elif directions[i] == "below":
-            if inclusive:
-                adata.obs.loc[
-                    (adata.obs[name] == 1)
-                    & (adata.obs[obs_cols[i]] >= thresholds[obs_cols[i]]),
-                    name,
-                ] = 0
-            else:
-                adata.obs.loc[
-                    (adata.obs[name] == 1)
-                    & (adata.obs[obs_cols[i]] > thresholds[obs_cols[i]]),
-                    name,
-                ] = 0
+        if isinstance(thresholds[obs_cols[i]], np.ndarray):
+
+            # set labels using second threshold for "above"
+            if directions[i] == "above":
+                if inclusive:
+                    adata.obs.loc[
+                        (adata.obs[name] == 1)
+                        & (adata.obs[obs_cols[i]] <= thresholds[obs_cols[i]][1]),
+                        name,
+                    ] = 0
+                else:
+                    adata.obs.loc[
+                        (adata.obs[name] == 1)
+                        & (adata.obs[obs_cols[i]] < thresholds[obs_cols[i]][1]),
+                        name,
+                    ] = 0
+            # set labels using first threshold for "below"
+            elif directions[i] == "below":
+                if inclusive:
+                    adata.obs.loc[
+                        (adata.obs[name] == 1)
+                        & (adata.obs[obs_cols[i]] >= thresholds[obs_cols[i]][0]),
+                        name,
+                    ] = 0
+                else:
+                    adata.obs.loc[
+                        (adata.obs[name] == 1)
+                        & (adata.obs[obs_cols[i]] > thresholds[obs_cols[i]][0]),
+                        name,
+                    ] = 0
+
+        else:
+            # if single threshold, just assign labels
+            if directions[i] == "above":
+                if inclusive:
+                    adata.obs.loc[
+                        (adata.obs[name] == 1)
+                        & (adata.obs[obs_cols[i]] <= thresholds[obs_cols[i]]),
+                        name,
+                    ] = 0
+                else:
+                    adata.obs.loc[
+                        (adata.obs[name] == 1)
+                        & (adata.obs[obs_cols[i]] < thresholds[obs_cols[i]]),
+                        name,
+                    ] = 0
+            elif directions[i] == "below":
+                if inclusive:
+                    adata.obs.loc[
+                        (adata.obs[name] == 1)
+                        & (adata.obs[obs_cols[i]] >= thresholds[obs_cols[i]]),
+                        name,
+                    ] = 0
+                else:
+                    adata.obs.loc[
+                        (adata.obs[name] == 1)
+                        & (adata.obs[obs_cols[i]] > thresholds[obs_cols[i]]),
+                        name,
+                    ] = 0
+
+    return adata
 
 
 def dropkick(
@@ -299,8 +359,8 @@ def dropkick(
     mito_names="^mt-|^MT-",
     n_ambient=10,
     n_hvgs=2000,
-    thresh_method="otsu",
-    metrics=["arcsinh_n_genes_by_counts", "pct_counts_ambient",],
+    metrics=["arcsinh_n_genes_by_counts", "pct_counts_ambient"],
+    thresh_methods=["multiotsu", "otsu"],
     directions=["above", "below"],
     alphas=[0.1],
     max_iter=1000,
@@ -314,18 +374,18 @@ def dropkick(
     Parameters:
         adata (anndata.AnnData): object containing unfiltered, raw scRNA-seq
             counts in .X layer
-        min_genes (int): threshold for minimum genes detected. Default 50.
-            Ignores all cells with less than min_genes (dropkick label = 0).
+        min_genes (int): threshold for minimum genes detected. Ignores all cells
+            with less than min_genes (dropkick label = 0).
         mito_names (str): substring encompassing mitochondrial gene names for
             calculation of mito expression
         n_ambient (int): number of ambient genes to call. top genes by cells.
-        n_hvgs (int or None): number of HVGs to calculate using Seurat method
+        n_hvgs (int or None): number of HVGs to calculate using Seurat method.
             if None, do not calculate HVGs
-        thresh_method (str): one of 'otsu' (default), 'li', or 'mean'
         metrics (list of str): name of column(s) to threshold from adata.obs
+        thresh_methods (list of str): one of 'otsu', 'multiotsu', 'li', or 'mean'
         directions (list of str): 'below' or 'above', indicating which
-            direction to keep (label=1)
-        alphas (tuple of int): alpha values to test using glmnet with n-fold
+            direction to keep (dropkick label = 1)
+        alphas (list of float): alpha values to test using glmnet with n-fold
             cross validation
         max_iter (int): number of iterations for glmnet optimization
         n_jobs (int): number of threads for cross validation by glmnet
@@ -333,7 +393,6 @@ def dropkick(
         verbose (bool): verbosity for glmnet training and warnings
 
     Returns:
-        adata_thresh (dict): dictionary of automated thresholds on heuristics
         rc (LogisticRegression): trained logistic regression classifier
 
         updates adata inplace to include 'train', 'dropkick_score', and
@@ -354,29 +413,37 @@ def dropkick(
         verbose=verbose,
     )
 
-    # 1) threshold chosen heuristics using automated method
+    # 1) threshold chosen heuristics using automated methods
     print("Thresholding on heuristics for training labels:\n\t{}".format(metrics))
-    adata_thresh = auto_thresh_obs(a, method=thresh_method, obs_cols=metrics)
+    # convert args to list
+    if isinstance(metrics, str):
+        metrics = [metrics]
+    if isinstance(thresh_methods, str):
+        thresh_methods = [thresh_methods]
+    adata_thresh = auto_thresh_obs(a, methods=thresh_methods, obs_cols=metrics)
 
     # 2) create labels from combination of thresholds
-    filter_thresh_obs(
+    # convert args to list
+    if isinstance(directions, str):
+        directions = [directions]
+    a = filter_thresh_obs(
         a,
         adata_thresh,
         obs_cols=metrics,
         directions=directions,
         inclusive=True,
         name="train",
+        verbose=verbose,
     )
 
     X = a[:, a.var.highly_variable].X.copy()  # X for testing
     y = a.obs["train"].copy(deep=True)  # final y is "train" labels from step 2
-    print("Training LogitNet with alphas: {}".format(alphas))
+    print("Training dropkick with alphas: {}".format(alphas))
 
     if len(alphas) > 1:
         # 3.1) cross-validation to choose alpha and lambda values
         cv_scores = {"rc": [], "lambda": [], "alpha": [], "score": []}  # dictionary o/p
         for alpha in alphas:
-            # print("Training LogitNet with alpha: {}".format(alpha), end="  ")
             rc = LogitNet(
                 alpha=alpha,
                 n_lambda=100,
@@ -428,15 +495,18 @@ def dropkick(
     for metric in metrics:
         adata.obs.loc[a.obs_names, metric] = a.obs[metric]
         adata.obs[metric].fillna(0, inplace=True)  # fill ignored cells with zeros
-    adata.var.loc[a.var.highly_variable, "dropkick_coef"] = rc_.coef_.squeeze()
+    # add dropkick coefficients to genes used in model (hvgs from `a`)
+    adata.var.loc[
+        a.var_names[a.var.highly_variable], "dropkick_coef"
+    ] = rc_.coef_.squeeze()
 
     # 5) save model hyperparameters in .uns
     adata.uns["dropkick_thresholds"] = adata_thresh
     adata.uns["dropkick_args"] = {
         "n_ambient": n_ambient,
         "n_hvgs": n_hvgs,
-        "thresh_method": thresh_method,
         "metrics": metrics,
+        "thresh_methods": thresh_methods,
         "directions": directions,
         "alphas": alphas,
         "chosen_alpha": alpha_,
@@ -479,14 +549,15 @@ def coef_inventory(adata, n=10):
     )
 
 
-def coef_plot(adata, show=True):
+def coef_plot(adata, save_to=None, verbose=True):
     """
     plot dropkick coefficient values and cross validation (CV) scores for tested values
     of lambda (lambda_path)
 
     Parameters:
         adata (anndata.AnnData): object generated from dropkick
-        show (bool): show plot or return object
+        save_to (str): path to .png file for saving figure; returns figure by default
+        verbose (bool): print updates to console
 
     Returns:
         plot of CV scores (mean +/- SEM) and coefficient values (coef_path) versus
@@ -503,10 +574,17 @@ def coef_plot(adata, show=True):
         alpha=0.5,
     )
     # plot top three genes by coefficient value
+    # get range of values for positioning text
+    val_range = (
+        adata.uns["dropkick_args"]["coef_path"].max()
+        - adata.uns["dropkick_args"]["coef_path"].min()
+    )
     [
         ax.text(
             x=np.log(adata.uns["dropkick_args"]["chosen_lambda"]),
-            y=adata.var.dropkick_coef.max() + 0.15 - 0.05 * x,
+            y=adata.var.dropkick_coef.max()
+            + (0.15 * val_range)
+            - (0.05 * val_range * x),
             s=" "
             + adata.var.loc[-adata.var.dropkick_coef.isna(), "dropkick_coef"]
             .nlargest(3)
@@ -520,7 +598,9 @@ def coef_plot(adata, show=True):
     [
         ax.text(
             x=np.log(adata.uns["dropkick_args"]["chosen_lambda"]),
-            y=adata.var.dropkick_coef.min() - 0.20 + 0.05 * x,
+            y=adata.var.dropkick_coef.min()
+            - (0.20 * val_range)
+            + (0.05 * val_range * x),
             s=" "
             + adata.var.loc[-adata.var.dropkick_coef.isna(), "dropkick_coef"]
             .nsmallest(3)
@@ -562,7 +642,137 @@ def coef_plot(adata, show=True):
     )
     plt.legend()
     fig.tight_layout()
-    if show:
-        plt.show()
+    if save_to is not None:
+        if verbose:
+            print("Saving coefficient plot to {}".format(save_to))
+        fig.savefig(save_to)
     else:
         return fig
+
+
+def score_plot(
+    adata,
+    metrics=["arcsinh_n_genes_by_counts", "pct_counts_ambient"],
+    save_to=None,
+    verbose=True,
+):
+    """
+    plot scatter of barcodes across two metrics, with points colored by dropkick_score.
+    show automated thresholding on metrics in adata.obs as output by auto_thresh_obs()
+
+    Parameters:
+        adata (anndata.AnnData): object containing dropkick-processed scRNA-seq data
+        metrics (list of str): name of metrics to plot scatter and histograms for
+        save_to (str): path to .png file for saving figure; returns figure by default
+        verbose (bool): print updates to console
+
+    Returns:
+        joint plot of dropkick_scores and metric distributions with
+            corresponding training thresholds
+    """
+    # initialize joint plot object
+    g = sns.jointplot(
+        x=adata.obs[metrics[0]],
+        y=adata.obs[metrics[1]],
+        space=0,
+        color="k",
+        marginal_kws=dict(bins=40),
+    )
+    # change to focus on scatter plot
+    g.ax_joint.cla()
+    plt.sca(g.ax_joint)
+    # set axes labels
+    plt.xlabel(metrics[0])
+    plt.ylabel(metrics[1])
+    # scatter plot, color by dropkick_score
+    points = plt.scatter(
+        x=adata.obs[metrics[0]],
+        y=adata.obs[metrics[1]],
+        c=adata.obs["dropkick_score"],
+        s=25,
+        cmap="coolwarm_r",
+        alpha=0.5,
+    )
+    # plot training thresholds on scatter
+    if metrics[0] in adata.uns["dropkick_thresholds"]:
+        if isinstance(adata.uns["dropkick_thresholds"][metrics[0]], np.ndarray):
+            [
+                plt.axvline(_x, linestyle="-", color="k", linewidth=2.5, alpha=0.7)
+                for _x in adata.uns["dropkick_thresholds"][metrics[0]]
+            ]
+        else:
+            plt.axvline(
+                adata.uns["dropkick_thresholds"][metrics[0]],
+                linestyle="-",
+                color="k",
+                linewidth=2.5,
+                alpha=0.7,
+            )
+    if metrics[1] in adata.uns["dropkick_thresholds"]:
+        if isinstance(adata.uns["dropkick_thresholds"][metrics[1]], np.ndarray):
+            [
+                plt.axhline(_x, linestyle="-", color="k", linewidth=2.5, alpha=0.7)
+                for _x in adata.uns["dropkick_thresholds"][metrics[1]]
+            ]
+        else:
+            plt.axhline(
+                adata.uns["dropkick_thresholds"][metrics[1]],
+                linestyle="-",
+                color="k",
+                linewidth=2.5,
+                alpha=0.7,
+            )
+    # change focus to x margin plot to continue threshold line
+    if metrics[0] in adata.uns["dropkick_thresholds"]:
+        plt.sca(g.ax_marg_x)
+        if isinstance(adata.uns["dropkick_thresholds"][metrics[0]], np.ndarray):
+            [
+                plt.axvline(_x, linestyle="-", color="k", linewidth=2.5, alpha=0.7)
+                for _x in adata.uns["dropkick_thresholds"][metrics[0]]
+            ]
+        else:
+            plt.axvline(
+                adata.uns["dropkick_thresholds"][metrics[0]],
+                linestyle="-",
+                color="k",
+                linewidth=2.5,
+                alpha=0.7,
+            )
+    # change focus to y margin plot to continue threshold line
+    if metrics[1] in adata.uns["dropkick_thresholds"]:
+        plt.sca(g.ax_marg_y)
+        if isinstance(adata.uns["dropkick_thresholds"][metrics[1]], np.ndarray):
+            [
+                plt.axhline(_x, linestyle="-", color="k", linewidth=2.5, alpha=0.7)
+                for _x in adata.uns["dropkick_thresholds"][metrics[1]]
+            ]
+        else:
+            plt.axhline(
+                adata.uns["dropkick_thresholds"][metrics[1]],
+                linestyle="-",
+                color="k",
+                linewidth=2.5,
+                alpha=0.7,
+            )
+    # add colorbar inside scatter axes
+    axins1 = inset_axes(
+        g.ax_joint,
+        width="40%",  # width = 40% of parent_bbox width
+        height="3%",  # height : 3%
+        loc="upper right",
+    )
+    cbar = plt.colorbar(
+        points,
+        cax=axins1,
+        drawedges=False,
+        label="dropkick_score",
+        orientation="horizontal",
+        ticks=[0.1, 0.5, 0.9],
+    )
+    cbar.solids.set_edgecolor("face")
+    if save_to is not None:
+        if verbose:
+            print("Saving score plot to {}".format(save_to))
+        g.savefig(save_to)
+    else:
+        return g
