@@ -115,7 +115,7 @@ class LogitNet(BaseEstimator):
         The set of coefficients for each value of lambda in lambda_path_.
 
     coef_ : array, shape (n_clases, n_features)
-        The coefficients corresponding to lamnda_best_.
+        The coefficients corresponding to lambda_best_.
 
     intercept_ : array, shape (n_classes,)
         The intercept corresponding to lambda_best_.
@@ -199,12 +199,14 @@ class LogitNet(BaseEstimator):
 
         Parameters
         ----------
-        X : array, shape (n_samples, n_features)
-            Input features
+        adata : anndata.AnnData, shape (n_samples, n_features)
+            Input features in .X
 
-        Y : array, shape (n_samples,)
+        y : array, shape (n_samples,)
             Target values
 
+        n_hvgs : int, number of highly-variable genes to feature-select
+
         sample_weight : array, shape (n_samples,)
             Optional weight vector for observations
 
@@ -217,14 +219,8 @@ class LogitNet(BaseEstimator):
         self : object
             Returns self.
         """
-        a = adata.copy()  # select training set for fold
-        a.X = a.layers["log1p_norm"].copy()  # use log1p-transformed counts to calc HVGs
-        sc.pp.highly_variable_genes(
-            a, n_top_genes=n_hvgs, flavor="seurat", inplace=True
-        )  # determine HVGs with Seurat method
-        X = adata.X[
-            :, a.var.highly_variable
-        ].copy()  # X becomes scaled counts for all cells in HVGs only
+        # X is scaled counts for all cells in HVGs only for first run
+        X = adata[:, adata.var.highly_variable].X.copy()
 
         X, y = check_X_y(X, y, accept_sparse="csr", ensure_min_samples=2)
         if sample_weight is None:
@@ -269,7 +265,7 @@ class LogitNet(BaseEstimator):
                 n_splits=self.n_splits, shuffle=True, random_state=self.random_state
             )
 
-            cv_scores = _score_lambda_path(
+            cv_scores, hvgs = _score_lambda_path(
                 self,
                 adata,
                 y,
@@ -380,7 +376,6 @@ class LogitNet(BaseEstimator):
         else:
             max_features = self.max_features
 
-        # for documentation on the glmnet function lognet, see doc.py
         if issparse(X):
             _x = csc_matrix(X, dtype=np.float64, copy=True)
 
